from empiric.core.contract import Contract
from starknet_py.net.client import Client
from starknet_py.net.client_models import Call
from starkware.starknet.public.abi import get_selector_from_name


class TransactionMixin:
    publisher_registry: Contract
    client: Client

    async def send_transaction(self, to_contract, selector_name, calldata) -> str:
        selector = get_selector_from_name(selector_name)
        return await self.send_transactions([Call(to_contract, selector, calldata)])

    async def send_transactions(self, calls) -> str:
        return hex(
            (await self.client.execute(calls, auto_estimate=True)).transaction_hash
        )



import os
from enum import IntEnum, unique
from typing import List, Literal

from empiric.core.utils import str_to_felt
from starknet_py.net.full_node_client import FullNodeClient

NETWORK = os.getenv("NETWORK") or "sharingan"

ADDRESS = int
HEX_STR = str

# Network Types
STAGING = "staging"
TESTNET = "testnet"
INTEGRATION = "integration"
MAINNET = "mainnet"
SCROLL_TESTNET = "scroll_testnet"
LINEA_TESTNET = "linea_testnet"
ERA_TESTNET = "era_testnet"
SHARINGAN = "sharingan"
PRAGMA_TESTNET = "pragma_testnet"

Network = Literal["staging", "testnet", "integration", "mainnet", "sharingan", "pragma_testnet"]

CHAIN_IDS = {
    INTEGRATION: 1536727068981429685321,
    SHARINGAN: 1536727068981429685321,
    TESTNET: 1536727068981429685321,
    MAINNET: 23448594291968334,
    SCROLL_TESTNET: 534353,
    LINEA_TESTNET: 59140,
    ERA_TESTNET: 280,
    PRAGMA_TESTNET: 8908953246943201047421899664489
}

STARKSCAN_URLS = {
    "mainnet": "https://starkscan.co",
    "testnet": "https://testnet.starkscan.co",
    "testnet2": "https://testnet-2.starkscan.co",
    "devnet": "https://devnet.starkscan.co",
    "sharingan": "https://starknet-madara.netlify.app/#/explorer/query",
    "pragma_testnet": "https://testnet.pragmaoracle.com/explorer",
}

if not os.getenv("RPC_KEY") and NETWORK in ["mainnet", "testnet", "testnet2"]:
    raise ValueError(f"RPC_KEY env variable is required when targeting {NETWORK}")
RPC_URLS = {
    "mainnet": f"https://starknet-mainnet.infura.io/v3/{os.getenv('RPC_KEY')}",
    "testnet": f"https://starknet-goerli.infura.io/v3/{os.getenv('RPC_KEY')}",
    "testnet2": f"https://starknet-goerli2.infura.io/v3/{os.getenv('RPC_KEY')}",
    "devnet": "http://127.0.0.1:5050/rpc",
    "sharingan": "https://sharingan.madara.zone",
    "pragma_testnet": "https://testnet.pragmaoracle.com/rpc"
}
RPC_CLIENT = FullNodeClient(node_url=RPC_URLS[NETWORK])


# aggregation mode enum
@unique
class AggregationMode(IntEnum):
    MEDIAN = 84959893733710  # str_to_felt("MEDIAN")


class Currency:
    id: int
    decimals: int
    is_abstract_currency: int
    starknet_address: int
    ethereum_address: int

    def __init__(
        self,
        id,
        decimals,
        is_abstract_currency,
        starknet_address=None,
        ethereum_address=None,
    ):
        if type(id) == str:
            id = str_to_felt(id)
        self.id = id

        self.decimals = decimals

        if type(is_abstract_currency) == bool:
            is_abstract_currency = int(is_abstract_currency)
        self.is_abstract_currency = is_abstract_currency

        if starknet_address is None:
            starknet_address = 0
        self.starknet_address = starknet_address

        if ethereum_address is None:
            ethereum_address = 0
        self.ethereum_address = ethereum_address

    def serialize(self) -> List[str]:
        return [
            self.id,
            self.decimals,
            self.is_abstract_currency,
            self.starknet_address,
            self.ethereum_address,
        ]

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "decimals": self.decimals,
            "is_abstract_currency": self.is_abstract_currency,
            "starknet_address": self.starknet_address,
            "ethereum_address": self.ethereum_address,
        }


class Pair:
    id: int
    quote_currency_id: int
    base_currency_id: int

    def __init__(self, id, quote_currency_id, base_currency_id):
        if type(id) == str:
            id = str_to_felt(id)
        self.id = id

        if type(quote_currency_id) == str:
            quote_currency_id = str_to_felt(quote_currency_id)
        self.quote_currency_id = quote_currency_id

        if type(base_currency_id) == str:
            base_currency_id = str_to_felt(base_currency_id)
        self.base_currency_id = base_currency_id

    def serialize(self) -> List[str]:
        return [self.id, self.quote_currency_id, self.base_currency_id]

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "quote_currency_id": self.quote_currency_id,
            "base_currency_id": self.base_currency_id,
        }



ORACLE_ABI = [
    {
        "members": [
            {"name": "base", "offset": 0, "type": "BaseEntry"},
            {"name": "key", "offset": 3, "type": "felt"},
            {"name": "value", "offset": 4, "type": "felt"},
        ],
        "name": "GenericEntry",
        "size": 5,
        "type": "struct",
    },
    {
        "members": [
            {"name": "timestamp", "offset": 0, "type": "felt"},
            {"name": "source", "offset": 1, "type": "felt"},
            {"name": "publisher", "offset": 2, "type": "felt"},
        ],
        "name": "BaseEntry",
        "size": 3,
        "type": "struct",
    },
    {
        "members": [
            {"name": "base", "offset": 0, "type": "BaseEntry"},
            {"name": "pair_id", "offset": 3, "type": "felt"},
            {"name": "price", "offset": 4, "type": "felt"},
            {"name": "volume", "offset": 5, "type": "felt"},
        ],
        "name": "SpotEntry",
        "size": 6,
        "type": "struct",
    },
    {
        "members": [
            {"name": "base", "offset": 0, "type": "BaseEntry"},
            {"name": "pair_id", "offset": 3, "type": "felt"},
            {"name": "price", "offset": 4, "type": "felt"},
            {"name": "expiry_timestamp", "offset": 5, "type": "felt"},
        ],
        "name": "FutureEntry",
        "size": 6,
        "type": "struct",
    },
    {
        "members": [
            {"name": "id", "offset": 0, "type": "felt"},
            {"name": "decimals", "offset": 1, "type": "felt"},
            {"name": "is_abstract_currency", "offset": 2, "type": "felt"},
            {"name": "starknet_address", "offset": 3, "type": "felt"},
            {"name": "ethereum_address", "offset": 4, "type": "felt"},
        ],
        "name": "Currency",
        "size": 5,
        "type": "struct",
    },
    {
        "members": [
            {"name": "id", "offset": 0, "type": "felt"},
            {"name": "quote_currency_id", "offset": 1, "type": "felt"},
            {"name": "base_currency_id", "offset": 2, "type": "felt"},
        ],
        "name": "Pair",
        "size": 3,
        "type": "struct",
    },
    {
        "members": [
            {"name": "timestamp", "offset": 0, "type": "felt"},
            {"name": "value", "offset": 1, "type": "felt"},
            {"name": "aggregation_mode", "offset": 2, "type": "felt"},
            {"name": "num_sources_aggregated", "offset": 3, "type": "felt"},
        ],
        "name": "Checkpoint",
        "size": 4,
        "type": "struct",
    },
    {
        "data": [
            {"name": "old_publisher_registry_address", "type": "felt"},
            {"name": "new_publisher_registry_address", "type": "felt"},
        ],
        "keys": [],
        "name": "UpdatedPublisherRegistryAddress",
        "type": "event",
    },
    {
        "data": [{"name": "new_entry", "type": "GenericEntry"}],
        "keys": [],
        "name": "SubmittedEntry",
        "type": "event",
    },
    {
        "data": [{"name": "new_entry", "type": "SpotEntry"}],
        "keys": [],
        "name": "SubmittedSpotEntry",
        "type": "event",
    },
    {
        "data": [{"name": "new_entry", "type": "FutureEntry"}],
        "keys": [],
        "name": "SubmittedFutureEntry",
        "type": "event",
    },
    {
        "data": [{"name": "currency", "type": "Currency"}],
        "keys": [],
        "name": "SubmittedCurrency",
        "type": "event",
    },
    {
        "data": [{"name": "currency", "type": "Currency"}],
        "keys": [],
        "name": "UpdatedCurrency",
        "type": "event",
    },
    {
        "data": [{"name": "pair", "type": "Pair"}],
        "keys": [],
        "name": "SubmittedPair",
        "type": "event",
    },
    {
        "data": [{"name": "pair_id", "type": "felt"}],
        "keys": [],
        "name": "CheckpointSpotEntry",
        "type": "event",
    },
    {
        "data": [{"name": "implementation", "type": "felt"}],
        "keys": [],
        "name": "Upgraded",
        "type": "event",
    },
    {
        "data": [
            {"name": "previousAdminAddress", "type": "felt"},
            {"name": "newAdminAddress", "type": "felt"},
        ],
        "keys": [],
        "name": "AdminAddressChanged",
        "type": "event",
    },
    {
        "inputs": [
            {"name": "proxy_admin", "type": "felt"},
            {"name": "publisher_registry_address", "type": "felt"},
            {"name": "currencies_len", "type": "felt"},
            {"name": "currencies", "type": "Currency*"},
            {"name": "pairs_len", "type": "felt"},
            {"name": "pairs", "type": "Pair*"},
        ],
        "name": "initializer",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "sources_len", "type": "felt"},
            {"name": "sources", "type": "felt*"},
        ],
        "name": "get_spot_entries_for_sources",
        "outputs": [
            {"name": "entries_len", "type": "felt"},
            {"name": "entries", "type": "SpotEntry*"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "pair_id", "type": "felt"}],
        "name": "get_spot_entries",
        "outputs": [
            {"name": "entries_len", "type": "felt"},
            {"name": "entries", "type": "SpotEntry*"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "source", "type": "felt"},
        ],
        "name": "get_spot_entry",
        "outputs": [{"name": "entry", "type": "SpotEntry"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "expiry_timestamp", "type": "felt"},
            {"name": "source", "type": "felt"},
        ],
        "name": "get_future_entry",
        "outputs": [{"name": "entry", "type": "FutureEntry"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "pair_id", "type": "felt"}],
        "name": "get_spot_median",
        "outputs": [
            {"name": "price", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "sources_len", "type": "felt"},
            {"name": "sources", "type": "felt*"},
        ],
        "name": "get_spot_median_for_sources",
        "outputs": [
            {"name": "price", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "aggregation_mode", "type": "felt"},
        ],
        "name": "get_spot",
        "outputs": [
            {"name": "price", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "aggregation_mode", "type": "felt"},
            {"name": "sources_len", "type": "felt"},
            {"name": "sources", "type": "felt*"},
        ],
        "name": "get_spot_for_sources",
        "outputs": [
            {"name": "price", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_publisher_registry_address",
        "outputs": [{"name": "publisher_registry_address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "pair_id", "type": "felt"}],
        "name": "get_spot_decimals",
        "outputs": [{"name": "decimals", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "key", "type": "felt"}],
        "name": "get_value",
        "outputs": [
            {"name": "value", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "base_currency_id", "type": "felt"},
            {"name": "quote_currency_id", "type": "felt"},
            {"name": "aggregation_mode", "type": "felt"},
        ],
        "name": "get_spot_with_USD_hop",
        "outputs": [
            {"name": "price", "type": "felt"},
            {"name": "decimals", "type": "felt"},
            {"name": "last_updated_timestamp", "type": "felt"},
            {"name": "num_sources_aggregated", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "new_entry", "type": "FutureEntry"}],
        "name": "publish_future_entry",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "new_entry", "type": "SpotEntry"}],
        "name": "publish_spot_entry",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "new_entry", "type": "GenericEntry"}],
        "name": "publish_entry",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "new_entries_len", "type": "felt"},
            {"name": "new_entries", "type": "GenericEntry*"},
        ],
        "name": "publish_entries",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "new_entries_len", "type": "felt"},
            {"name": "new_entries", "type": "FutureEntry*"},
        ],
        "name": "publish_future_entries",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "new_entries_len", "type": "felt"},
            {"name": "new_entries", "type": "SpotEntry*"},
        ],
        "name": "publish_spot_entries",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "publisher_registry_address", "type": "felt"}],
        "name": "update_publisher_registry_address",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "currency", "type": "Currency"}],
        "name": "add_currency",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "currency", "type": "Currency"}],
        "name": "update_currency",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "pair", "type": "Pair"}],
        "name": "add_pair",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "key", "type": "felt"}],
        "name": "get_latest_checkpoint_index",
        "outputs": [{"name": "latest", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "key", "type": "felt"}, {"name": "index", "type": "felt"}],
        "name": "get_checkpoint",
        "outputs": [{"name": "checkpoint", "type": "Checkpoint"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_sources_threshold",
        "outputs": [{"name": "threshold", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "new_implementation", "type": "felt"}],
        "name": "upgrade",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "new_admin_address", "type": "felt"}],
        "name": "set_admin_address",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_implementation_hash",
        "outputs": [{"name": "address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_admin_address",
        "outputs": [{"name": "admin_address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "aggregation_mode", "type": "felt"},
        ],
        "name": "set_checkpoint",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_ids_len", "type": "felt"},
            {"name": "pair_ids", "type": "felt*"},
            {"name": "aggregation_mode", "type": "felt"},
        ],
        "name": "set_checkpoints",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "pair_id", "type": "felt"},
            {"name": "timestamp", "type": "felt"},
        ],
        "name": "get_last_spot_checkpoint_before",
        "outputs": [
            {"name": "checkpoint", "type": "Checkpoint"},
            {"name": "idx", "type": "felt"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "threshold", "type": "felt"}],
        "name": "set_sources_threshold",
        "outputs": [],
        "type": "function",
    },
]



import os
import time
from typing import List

from empiric.core.entry import SpotEntry
from empiric.core.types import CHAIN_IDS, GATEWAY_URLS, LINEA_TESTNET, SCROLL_TESTNET, ERA_TESTNET
from web3 import Web3
from web3.middleware import geth_poa_middleware

ORACLE_ABI = [
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "uint64", "name": "timestamp", "type": "uint64"},
                    {"internalType": "uint128", "name": "value", "type": "uint128"},
                    {
                        "internalType": "enum IOracle.AggregationMode",
                        "name": "aggregationMode",
                        "type": "uint8",
                    },
                    {
                        "internalType": "uint8",
                        "name": "numSourcesAggregated",
                        "type": "uint8",
                    },
                ],
                "indexed": False,
                "internalType": "struct IOracle.Checkpoint",
                "name": "cp",
                "type": "tuple",
            }
        ],
        "name": "CheckpointSpotEntry",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "version",
                "type": "uint8",
            }
        ],
        "name": "Initialized",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "previousOwner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newOwner",
                "type": "address",
            },
        ],
        "name": "OwnershipTransferred",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {"internalType": "uint256", "name": "decimals", "type": "uint256"},
                    {
                        "internalType": "bool",
                        "name": "isAbstractCurrency",
                        "type": "bool",
                    },
                    {
                        "internalType": "address",
                        "name": "ethereumAddress",
                        "type": "address",
                    },
                ],
                "indexed": False,
                "internalType": "struct ICurrencyManager.Currency",
                "name": "currency",
                "type": "tuple",
            }
        ],
        "name": "SubmittedCurrency",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {
                        "internalType": "bytes32",
                        "name": "quoteCurrencyId",
                        "type": "bytes32",
                    },
                    {
                        "internalType": "bytes32",
                        "name": "baseCurrencyId",
                        "type": "bytes32",
                    },
                ],
                "indexed": False,
                "internalType": "struct ICurrencyManager.Pair",
                "name": "pair",
                "type": "tuple",
            }
        ],
        "name": "SubmittedPair",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "timestamp",
                                "type": "uint256",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "source",
                                "type": "bytes32",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "publisher",
                                "type": "bytes32",
                            },
                        ],
                        "internalType": "struct IOracle.BaseEntry",
                        "name": "base",
                        "type": "tuple",
                    },
                    {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
                    {"internalType": "uint256", "name": "price", "type": "uint256"},
                    {"internalType": "uint256", "name": "volume", "type": "uint256"},
                ],
                "indexed": False,
                "internalType": "struct IOracle.SpotEntry",
                "name": "newEntry",
                "type": "tuple",
            }
        ],
        "name": "SubmittedSpotEntry",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {"internalType": "uint256", "name": "decimals", "type": "uint256"},
                    {
                        "internalType": "bool",
                        "name": "isAbstractCurrency",
                        "type": "bool",
                    },
                    {
                        "internalType": "address",
                        "name": "ethereumAddress",
                        "type": "address",
                    },
                ],
                "indexed": False,
                "internalType": "struct ICurrencyManager.Currency",
                "name": "currency",
                "type": "tuple",
            }
        ],
        "name": "UpdatedCurrency",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "oldPublisherRegistryAddress",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "newPublisherRegistryAddress",
                "type": "address",
            },
        ],
        "name": "UpdatedPublisherRegistryAddress",
        "type": "event",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {"internalType": "uint256", "name": "decimals", "type": "uint256"},
                    {
                        "internalType": "bool",
                        "name": "isAbstractCurrency",
                        "type": "bool",
                    },
                    {
                        "internalType": "address",
                        "name": "ethereumAddress",
                        "type": "address",
                    },
                ],
                "internalType": "struct ICurrencyManager.Currency",
                "name": "currency",
                "type": "tuple",
            }
        ],
        "name": "addCurrency",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {
                        "internalType": "bytes32",
                        "name": "quoteCurrencyId",
                        "type": "bytes32",
                    },
                    {
                        "internalType": "bytes32",
                        "name": "baseCurrencyId",
                        "type": "bytes32",
                    },
                ],
                "internalType": "struct ICurrencyManager.Pair",
                "name": "pair",
                "type": "tuple",
            }
        ],
        "name": "addPair",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "checkpointIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "checkpoints",
        "outputs": [
            {"internalType": "uint64", "name": "timestamp", "type": "uint64"},
            {"internalType": "uint128", "name": "value", "type": "uint128"},
            {
                "internalType": "enum IOracle.AggregationMode",
                "name": "aggregationMode",
                "type": "uint8",
            },
            {"internalType": "uint8", "name": "numSourcesAggregated", "type": "uint8"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "currencies",
        "outputs": [
            {"internalType": "bytes32", "name": "id", "type": "bytes32"},
            {"internalType": "uint256", "name": "decimals", "type": "uint256"},
            {"internalType": "bool", "name": "isAbstractCurrency", "type": "bool"},
            {"internalType": "address", "name": "ethereumAddress", "type": "address"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "oracleSourcesStorage",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
        ],
        "name": "pairIdStorage",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "pairs",
        "outputs": [
            {"internalType": "bytes32", "name": "id", "type": "bytes32"},
            {"internalType": "bytes32", "name": "quoteCurrencyId", "type": "bytes32"},
            {"internalType": "bytes32", "name": "baseCurrencyId", "type": "bytes32"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "publisherRegistry",
        "outputs": [
            {
                "internalType": "contract IPublisherRegistry",
                "name": "",
                "type": "address",
            }
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "renounceOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
        ],
        "name": "spotEntryStorage",
        "outputs": [
            {"internalType": "uint128", "name": "timestamp", "type": "uint128"},
            {"internalType": "bytes16", "name": "pairId", "type": "bytes16"},
            {"internalType": "uint128", "name": "price", "type": "uint128"},
            {"internalType": "uint128", "name": "volume", "type": "uint128"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "newOwner", "type": "address"}],
        "name": "transferOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {"internalType": "uint256", "name": "decimals", "type": "uint256"},
                    {
                        "internalType": "bool",
                        "name": "isAbstractCurrency",
                        "type": "bool",
                    },
                    {
                        "internalType": "address",
                        "name": "ethereumAddress",
                        "type": "address",
                    },
                ],
                "internalType": "struct ICurrencyManager.Currency",
                "name": "currency",
                "type": "tuple",
            }
        ],
        "name": "updateCurrency",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_publisherRegistry",
                "type": "address",
            },
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {"internalType": "uint256", "name": "decimals", "type": "uint256"},
                    {
                        "internalType": "bool",
                        "name": "isAbstractCurrency",
                        "type": "bool",
                    },
                    {
                        "internalType": "address",
                        "name": "ethereumAddress",
                        "type": "address",
                    },
                ],
                "internalType": "struct ICurrencyManager.Currency[]",
                "name": "_currencies",
                "type": "tuple[]",
            },
            {
                "components": [
                    {"internalType": "bytes32", "name": "id", "type": "bytes32"},
                    {
                        "internalType": "bytes32",
                        "name": "quoteCurrencyId",
                        "type": "bytes32",
                    },
                    {
                        "internalType": "bytes32",
                        "name": "baseCurrencyId",
                        "type": "bytes32",
                    },
                ],
                "internalType": "struct ICurrencyManager.Pair[]",
                "name": "_pairs",
                "type": "tuple[]",
            },
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "threshold", "type": "uint256"}],
        "name": "setSourcesThreshold",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "contract IPublisherRegistry",
                "name": "newPublisherRegistryAddress",
                "type": "address",
            }
        ],
        "name": "updatePublisherRegistryAddress",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "timestamp",
                                "type": "uint256",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "source",
                                "type": "bytes32",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "publisher",
                                "type": "bytes32",
                            },
                        ],
                        "internalType": "struct IOracle.BaseEntry",
                        "name": "base",
                        "type": "tuple",
                    },
                    {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
                    {"internalType": "uint256", "name": "price", "type": "uint256"},
                    {"internalType": "uint256", "name": "volume", "type": "uint256"},
                ],
                "internalType": "struct IOracle.SpotEntry",
                "name": "spotEntry",
                "type": "tuple",
            }
        ],
        "name": "publishSpotEntry",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
            {
                "internalType": "enum IOracle.AggregationMode",
                "name": "aggregationMode",
                "type": "uint8",
            },
        ],
        "name": "setCheckpoint",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32[]", "name": "pairIds", "type": "bytes32[]"},
            {
                "internalType": "enum IOracle.AggregationMode",
                "name": "aggregationMode",
                "type": "uint8",
            },
        ],
        "name": "setCheckpoints",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "timestamp",
                                "type": "uint256",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "source",
                                "type": "bytes32",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "publisher",
                                "type": "bytes32",
                            },
                        ],
                        "internalType": "struct IOracle.BaseEntry",
                        "name": "base",
                        "type": "tuple",
                    },
                    {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
                    {"internalType": "uint256", "name": "price", "type": "uint256"},
                    {"internalType": "uint256", "name": "volume", "type": "uint256"},
                ],
                "internalType": "struct IOracle.SpotEntry[]",
                "name": "spotEntries",
                "type": "tuple[]",
            }
        ],
        "name": "publishSpotEntries",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
            {
                "internalType": "enum IOracle.AggregationMode",
                "name": "",
                "type": "uint8",
            },
            {"internalType": "bytes32[]", "name": "sources", "type": "bytes32[]"},
        ],
        "name": "getSpot",
        "outputs": [
            {"internalType": "uint256", "name": "price", "type": "uint256"},
            {"internalType": "uint256", "name": "decimals", "type": "uint256"},
            {
                "internalType": "uint256",
                "name": "lastUpdatedTimestamp",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "numSourcesAggregated",
                "type": "uint256",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
            {"internalType": "bytes32[]", "name": "sources", "type": "bytes32[]"},
        ],
        "name": "getSpotEntries",
        "outputs": [
            {
                "components": [
                    {"internalType": "uint128", "name": "timestamp", "type": "uint128"},
                    {"internalType": "bytes16", "name": "pairId", "type": "bytes16"},
                    {"internalType": "uint128", "name": "price", "type": "uint128"},
                    {"internalType": "uint128", "name": "volume", "type": "uint128"},
                ],
                "internalType": "struct IOracle.SpotEntryStorage[]",
                "name": "entries",
                "type": "tuple[]",
            },
            {
                "internalType": "uint256",
                "name": "lastUpdatedTimestamp",
                "type": "uint256",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
]

ORACLE_ADDRESS = {
    LINEA_TESTNET: "0xa2c8B12F06c27F235F1732A95cb86667476951Ac",
    SCROLL_TESTNET: "0xbb91Ed469258069Ff2590CA11E1800DE05Bf6Ec7",
    ERA_TESTNET: "0x807dEEA2a8BA552b13C418F1E9Ac8a12af77D1B1",
}


class EvmHelper:
    def __init__(
        self,
        publisher,
        sender_address,
        private_key,
        network_name,
    ):
        self.w3 = Web3(Web3.HTTPProvider(GATEWAY_URLS[network_name]))

        # The following middleware is required for POA chains (polygon, bnb, consensys zkevm)
        # See here for why https://web3py.readthedocs.io/en/v5/middleware.html?highlight=geth_poa_middleware#why-is-geth-poa-middleware-necessary
        self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        
        self.chain_id = CHAIN_IDS[network_name]
        self.oracle = self.w3.eth.contract(
            address=ORACLE_ADDRESS[network_name],
            abi=ORACLE_ABI,
        )
        self.publisher = publisher
        self.sender = sender_address or os.environ["SENDER_ADDRESS"]
        self.private_key = private_key or os.environ["PRIVATE_KEY"]

    def publish_spot_entry(
        self,
        pair,
        price,
        source,
        volume=0,
        gas_price=int(1e8),
        gas=int(1e6)
    ):
        nonce = self.w3.eth.getTransactionCount(self.sender)
        txn = self.oracle.functions.publishSpotEntry(
            {
                "base": {
                    "timestamp": int(time.time()),
                    "source": source,
                    "publisher": self.publisher,
                },
                "pairId": b"ETH/USD",
                "price": price,
                "volume": volume,
            }
        ).buildTransaction(
            {
                "nonce": nonce,
                "gasPrice": gas_price,
                "gasPrice": gas,
                "chainId": self.chain_id,
                "from": self.sender,
            }
        )
        print(txn)
        signed_txn = self.w3.eth.account.signTransaction(
            txn, private_key=self.private_key
        )
        self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        return signed_txn.hash.hex()

    def publish_spot_entries(self, spot_entries: List[SpotEntry], gas_price=int(1e8), gas=int(1e6)):
        serialized_spot_entries = SpotEntry.serialize_entries_evm(spot_entries)
        nonce = self.w3.eth.get_transaction_count(self.sender)

        # Transaction params
        # If it's a legacy transaction, we need to set the gasPrice and gas
        build_params = {
                "nonce": nonce,
                "chainId": self.chain_id,
                "from": self.sender,
        }
        if gas_price > 0:
            build_params["gasPrice"] = gas_price
        if gas > 0:
            build_params["gas"] = gas

        transaction = self.oracle.functions.publishSpotEntries(
            serialized_spot_entries
        ).build_transaction(
            build_params
        )

        signed_tx = self.w3.eth.account.sign_transaction(transaction, self.private_key)

        txn_hash = self.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
        txn_receipt = self.w3.eth.wait_for_transaction_receipt(txn_hash)

        return txn_hash.hex()



import asyncio
import logging
import time
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class CoinbaseFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://api.coinbase.com/v2/exchange-rates?currency="
    SOURCE: str = "COINBASE"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> Union[SpotEntry, PublisherFetchError]:
        currency = asset["pair"][1]

        async with session.get(self.BASE_URL + currency) as resp:
            result = await resp.json()
            return self._construct(asset, result)

    def _fetch_pair_sync(
        self, asset: EmpiricSpotAsset
    ) -> Union[SpotEntry, PublisherFetchError]:
        currency = asset["pair"][1]

        resp = requests.get(self.BASE_URL + currency)
        resp.raise_for_status()

        result = resp.json()
        return self._construct(asset, result)

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Coinbase for non-spot asset {asset}")
                continue

            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    def fetch_sync(self) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Coinbase for non-spot asset {asset}")
                continue

            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        pair_id = currency_pair_to_pair_id(*pair)

        if pair[0] in result["data"]["rates"]:
            rate = float(result["data"]["rates"][pair[0]])
            price = 1 / rate
            price_int = int(price * (10 ** asset["decimals"]))
            timestamp = int(time.time())

            logger.info(f"Fetched price {price} for {pair_id} from Coinbase")

            return SpotEntry(
                pair_id=pair_id,
                price=price_int,
                timestamp=timestamp,
                source=self.SOURCE,
                publisher=self.publisher,
            )

        return PublisherFetchError(f"No entry found for {pair_id} from Coinbase")



PUBLISHER_REGISTRY_ABI = [
    {
        "data": [
            {"name": "publisher", "type": "felt"},
            {"name": "publisher_address", "type": "felt"},
        ],
        "keys": [],
        "name": "RegisteredPublisher",
        "type": "event",
    },
    {
        "data": [
            {"name": "publisher", "type": "felt"},
            {"name": "old_publisher_address", "type": "felt"},
            {"name": "new_publisher_address", "type": "felt"},
        ],
        "keys": [],
        "name": "UpdatedPublisherAddress",
        "type": "event",
    },
    {
        "data": [
            {"name": "old_admin_address", "type": "felt"},
            {"name": "new_admin_address", "type": "felt"},
        ],
        "keys": [],
        "name": "UpdatedAdminAddress",
        "type": "event",
    },
    {
        "inputs": [{"name": "admin_address", "type": "felt"}],
        "name": "constructor",
        "outputs": [],
        "type": "constructor",
    },
    {
        "inputs": [],
        "name": "get_admin_address",
        "outputs": [{"name": "admin_address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "publisher", "type": "felt"}],
        "name": "get_publisher_address",
        "outputs": [{"name": "publisher_address", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "get_all_publishers",
        "outputs": [
            {"name": "publishers_len", "type": "felt"},
            {"name": "publishers", "type": "felt*"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "publisher", "type": "felt"}],
        "name": "get_publisher_sources",
        "outputs": [
            {"name": "sources_len", "type": "felt"},
            {"name": "sources", "type": "felt*"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "source", "type": "felt"},
        ],
        "name": "can_publish_source",
        "outputs": [{"name": "is_valid", "type": "felt"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"name": "new_address", "type": "felt"}],
        "name": "set_admin_address",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "publisher_address", "type": "felt"},
        ],
        "name": "add_publisher",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [{"name": "publisher", "type": "felt"}],
        "name": "remove_publisher",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "new_publisher_address", "type": "felt"},
        ],
        "name": "update_publisher_address",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "source", "type": "felt"},
        ],
        "name": "add_source_for_publisher",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "sources_len", "type": "felt"},
            {"name": "sources", "type": "felt*"},
        ],
        "name": "add_sources_for_publisher",
        "outputs": [],
        "type": "function",
    },
    {
        "inputs": [
            {"name": "publisher", "type": "felt"},
            {"name": "source", "type": "felt"},
        ],
        "name": "remove_source_for_publisher",
        "outputs": [],
        "type": "function",
    },
]



import asyncio
from typing import Callable, Optional

from starknet_py.contract import Contract as StarknetContract
from starknet_py.contract import ContractFunction, InvokeResult
from starknet_py.net.client_models import SentTransactionResponse, TransactionStatus
from starknet_py.transaction_errors import (
    TransactionFailedError,
    TransactionNotReceivedError,
    TransactionRejectedError,
)


class Contract(StarknetContract):
    def __getattr__(self, attr):
        if attr in self._functions:
            return self._functions[attr]
        elif attr in dir(self):
            return getattr(self, attr)
        else:
            raise AttributeError("Invalid Attribute")


async def invoke_(
    self,
    *args,
    max_fee: Optional[int] = None,
    auto_estimate: bool = False,
    callback: Optional[Callable[[SentTransactionResponse], None]] = None,
    **kwargs,
) -> InvokeResult:
    """
    Allows for a callback in the invocation of a contract method.
    This is useful for tracking the nonce changes
    """
    prepared_call = self.prepare(*args, **kwargs)

    # transfer ownership to the prepared call
    self = prepared_call
    if max_fee is not None:
        self.max_fee = max_fee

    transaction = await self._account.sign_invoke_transaction(
        calls=self,
        max_fee=self.max_fee,
        auto_estimate=auto_estimate,
    )
    response = await self._client.send_transaction(transaction)
    if callback:
        await callback(transaction.nonce, response.transaction_hash)

    invoke_result = InvokeResult(
        hash=response.transaction_hash,
        _client=self._client,
        contract=self._contract_data,
        invoke_transaction=transaction,
    )

    # don't return invoke result until it is received or errors
    await wait_for_received(self._client, invoke_result.hash)

    return invoke_result


async def wait_for_received(
    client,
    tx_hash,
    check_interval=6,
) -> (int, TransactionStatus):
    # pylint: disable=too-many-branches
    """
    Awaits for transaction to get accepted or at least pending by polling its status

    :param client: Account client
    :param tx_hash: Transaction's hash
    :param check_interval: Defines interval between checks
    :return: Tuple containing block number and transaction status
    """
    if check_interval <= 0:
        raise ValueError("check_interval has to bigger than 0.")

    first_run = True
    try:
        while True:
            status = TransactionStatus.NOT_RECEIVED
            try:
                result = await client.get_transaction_receipt(tx_hash=tx_hash)
                status = result.status
            except:
                continue

            if status in (
                TransactionStatus.ACCEPTED_ON_L1,
                TransactionStatus.ACCEPTED_ON_L2,
            ):
                return result.block_number, status
            if status == TransactionStatus.PENDING:
                return result.block_number, status
            elif status == TransactionStatus.REJECTED:
                raise TransactionRejectedError(
                    message=result.rejection_reason,
                )
            elif status == TransactionStatus.NOT_RECEIVED:
                if not first_run:
                    raise TransactionNotReceivedError()
            elif status != TransactionStatus.RECEIVED:
                # This will never get executed with current possible transactions statuses
                raise TransactionFailedError(
                    message=result.rejection_reason,
                )
            elif status == TransactionStatus.RECEIVED:
                return 0, 0

            first_run = False
            await asyncio.sleep(check_interval)
    except asyncio.CancelledError as exc:
        raise TransactionNotReceivedError from exc


# patch contract function to use new invoke function
ContractFunction.invoke = invoke_



import asyncio
import logging
from typing import List, Union

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset, EmpiricSpotAsset
from empiric.publisher.types import PublisherFetchError, PublisherInterfaceT

logger = logging.getLogger(__name__)


class BitstampFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://www.bitstamp.net/api/v2/ticker"
    SOURCE: str = "BITSTAMP"
    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def _fetch_pair(
        self, asset: EmpiricSpotAsset, session: ClientSession
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}/{pair[0].lower()}{pair[1].lower()}"
        async with session.get(url) as resp:
            if resp.status == 404:
                return PublisherFetchError(
                    f"No data found for {'/'.join(pair)} from Bitstamp"
                )
            return self._construct(asset, await resp.json())

    def _fetch_pair_sync(
        self, asset: EmpiricSpotAsset
    ) -> Union[SpotEntry, PublisherFetchError]:
        pair = asset["pair"]
        url = f"{self.BASE_URL}/{pair[0].lower()}{pair[1].lower()}"
        resp = requests.get(url)
        if resp.status_code == 404:
            return PublisherFetchError(
                f"No data found for {'/'.join(pair)} from Bitstamp"
            )
        return self._construct(asset, resp.json())

    async def fetch(
        self, session: ClientSession
    ) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Bitstamp for non-spot asset {asset}")
                continue
            entries.append(asyncio.ensure_future(self._fetch_pair(asset, session)))
        return await asyncio.gather(*entries, return_exceptions=True)

    async def fetch_sync(self) -> List[Union[SpotEntry, PublisherFetchError]]:
        entries = []
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Bitstamp for non-spot asset {asset}")
                continue
            entries.append(self._fetch_pair_sync(asset))
        return entries

    def _construct(self, asset, result) -> SpotEntry:
        pair = asset["pair"]

        timestamp = int(result["timestamp"])
        price = float(result["last"])
        price_int = int(price * (10 ** asset["decimals"]))
        pair_id = currency_pair_to_pair_id(*pair)

        logger.info(f"Fetched price {price} for {'/'.join(pair)} from Bitstamp")

        return SpotEntry(
            pair_id=pair_id,
            price=price_int,
            timestamp=timestamp,
            source=self.SOURCE,
            publisher=self.publisher,
        )



import logging

logger = logging.getLogger(__name__)


def str_to_felt(text):
    if text.upper() != text:
        logger.warning(f"Converting lower to uppercase for str_to_felt: {text}")
        text = text.upper()
    b_text = bytes(text, "utf-8")
    return int.from_bytes(b_text, "big")


def felt_to_str(felt):
    num_bytes = (felt.bit_length() + 7) // 8
    bytes = felt.to_bytes(num_bytes, "big")
    return bytes.decode("utf-8")


def currency_pair_to_pair_id(quote, base):
    return f"{quote}/{base}".upper()


def log_entry(entry, logger=logger):
    logger.info(f"Entry: {entry.serialize()}")


def pair_id_for_asset(asset):
    pair_id = (
        asset["key"] if "key" in asset else currency_pair_to_pair_id(*asset["pair"])
    )
    return pair_id


def key_for_asset(asset):
    return asset["key"] if "key" in asset else currency_pair_to_pair_id(*asset["pair"])



from typing import List

from empiric.core.contract import Contract
from empiric.core.utils import str_to_felt
from starknet_py.contract import InvokeResult
from starknet_py.net.account.account import Account
from starknet_py.net.client import Client


class PublisherRegistryMixin:
    client: Client
    account: Account
    publisher_registry: Contract

    async def get_all_publishers(self) -> List[str]:
        result = await self.publisher_registry.get_all_publishers.call()
        return result.publishers

    async def get_publisher_address(self, publisher) -> str:
        result = await self.publisher_registry.get_publisher_address.call(publisher)
        return result.publisher_address

    async def get_publisher_sources(self, publisher) -> List[str]:
        result = await self.publisher_registry.get_publisher_sources.call(publisher)
        return result.sources

    async def add_publisher(
        self, publisher: str, publisher_address: int, max_fee=int(1e16)
    ) -> InvokeResult:
        invocation = await self.publisher_registry.add_publisher.invoke(
            str_to_felt(publisher),
            publisher_address,
            max_fee=max_fee,
        )
        return invocation

    async def add_source_for_publisher(
        self, publisher: str, source: str, max_fee=int(1e16)
    ) -> InvokeResult:
        invocation = await self.publisher_registry.add_source_for_publisher.invoke(
            str_to_felt(publisher),
            str_to_felt(source),
            max_fee=max_fee,
        )
        return invocation

    async def add_sources_for_publisher(
        self, publisher: str, sources: List[str], max_fee=int(1e16)
    ) -> InvokeResult:
        invocation = await self.publisher_registry.add_sources_for_publisher.invoke(
            str_to_felt(publisher),
            [str_to_felt(source) for source in sources],
            max_fee=max_fee,
        )
        return invocation

    async def update_publisher_address(
        self, publisher: str, publisher_address: int, max_fee=int(1e16)
    ) -> InvokeResult:
        invocation = await self.publisher_registry.update_publisher_address.invoke(
            str_to_felt(publisher),
            publisher_address,
            max_fee=max_fee,
        )
        return invocation



from empiric.core.abis.oracle import ORACLE_ABI
from empiric.core.abis.proxy import PROXY_ABI
from empiric.core.abis.publisher_registry import PUBLISHER_REGISTRY_ABI
from empiric.core.abis.randomness import RANDOMNESS_ABI
from empiric.core.abis.summary_stats import SUMMARY_STATS_ABI



from typing import Dict, List, Tuple, Union

from empiric.core.utils import key_for_asset
from typing_extensions import TypedDict


class EmpiricSpotAsset(TypedDict):
    type: str
    pair: Tuple[str, str]
    decimals: int


class EmpiricOnchainDetail(TypedDict):
    asset_name: str
    asset_address: str
    metric: str


class EmpiricOnchainAsset(TypedDict):
    type: str
    source: str
    key: str
    detail: EmpiricOnchainDetail
    decimals: int


EmpiricAsset = Union[EmpiricSpotAsset, EmpiricOnchainAsset]


EMPIRIC_ALL_ASSETS: List[EmpiricAsset] = [
    {"type": "SPOT", "pair": ("BTC", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("WBTC", "BTC"), "decimals": 8},
    {"type": "SPOT", "pair": ("WBTC", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("BTC", "EUR"), "decimals": 8},
    {"type": "SPOT", "pair": ("ETH", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("SOL", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("AVAX", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("DOGE", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("SHIB", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("TEMP", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("DAI", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("USDT", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("USDC", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("TUSD", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("BUSD", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("ETH", "MXN"), "decimals": 8},
    {"type": "SPOT", "pair": ("BNB", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("ADA", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("XRP", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("MATIC", "USD"), "decimals": 8},
    {"type": "SPOT", "pair": ("AAVE", "USD"), "decimals": 8},
    {"type": "FUTURE", "pair": ("BTC", "USD"), "decimals": 8},
    {"type": "FUTURE", "pair": ("ETH", "USD"), "decimals": 8},
    {
        "type": "ONCHAIN",
        "source": "AAVE",
        "key": "AAVE-ON-BORROW",
        "detail": {
            "asset_name": "USD Coin",
            "asset_address": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb480xb53c1a33016b2dc2ff3653530bff1848a515c8c5",
            "metric": "variableBorrowRate",
        },
        "decimals": 18,
    },
]

_EMPIRIC_ASSET_BY_KEY: Dict[str, EmpiricSpotAsset] = {
    key_for_asset(asset): asset
    for asset in EMPIRIC_ALL_ASSETS
    if asset["type"] == "SPOT"
}


def get_spot_asset_spec_for_pair_id(pair_id: str) -> EmpiricSpotAsset:
    if pair_id not in _EMPIRIC_ASSET_BY_KEY:
        raise ValueError(f"Pair ID not found: {pair_id}")
    return _EMPIRIC_ASSET_BY_KEY[pair_id]
